/*
 * Copyright (c) 1990-1994 Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by the Computer Systems
 *	Engineering Group at Lawrence Berkeley Laboratory.
 * 4. Neither the name of the University nor of the Laboratory may be used
 *    to endorse or promote products derived from this software without
 *    specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */

#ifndef lint
static char rcsid[] =
    "@(#) $Header: /n/olympus/opt/users/CVS/ns-1.1-master/blue.cc,v 1.1 1998/10/29 20:37:53 wuchang Exp $ (LBL)";
#endif

#include <math.h>
#include <string.h>
#include <sys/types.h>

#include "link.h"
#include "Tcl.h"
#include "packet.h"
#include "trace.h"
#include "flow.h"
#include "random.h"
#include "scheduler.h"

extern "C" { int strcasecmp(const char*, const char*); }

class BlueQueue : public Link {
 public:	
	BlueQueue();
	inline int qlen() { return(q_.length()); }

 protected:
	virtual void reset();
	virtual void plot();
	void plot1(int qlen);
	void pmark_plot(int method);
	virtual int drop_early(Packet* pkt);
	virtual void enque(Packet* pkt);
	virtual Packet* deque();
	void increment_pmark(int how);
	void decrement_pmark(int how);
	Queue q_;
	/*
	 * Static state.
	 */
	int bytes_;	/* bytes or packets? */
	int dummy_;	/* */
	int setbit_;
	int mean_pktsize_;
	double decrement_;
	double increment_;
	double iholdtime_;
	double dholdtime_;
	int dalgorithm_;
	int ialgorithm_;

	/*
	 * Dynamic state.
	 */
	int idle_;
	double idletime_;
	double ptc_;	/* packet time constant in packets/second */
	double ifreezetime_;
	double dfreezetime_;
	double pmark_;
};

static class BlueMatcher : public Matcher {
public:
	BlueMatcher() : Matcher("link") {}
	TclObject* match(const char* id) {
		if (strcasecmp(id, "blue") == 0)
			return (new BlueQueue);
		return (0);
	}
} matcher_blue;

BlueQueue::BlueQueue()
{
	Tcl& tcl = Tcl::instance();

	link_bool("ns_blue", "bytes", &bytes_ , 0);
	link_bool("ns_blue", "setbit", &setbit_ , 0);
	link_real("ns_blue", "decrement", &decrement_ , 0);
	link_real("ns_blue", "increment", &increment_ , 0);
	link_time("ns_blue", "dhold-time", &dholdtime_ , 0);
	link_time("ns_blue", "ihold-time", &iholdtime_ , 0);

	// link_int("ns_blue", "dummy", &dummy_ , 0);
	link_int("ns_blue", "dalgorithm", &dalgorithm_ , 0);
	link_int("ns_blue", "ialgorithm", &ialgorithm_ , 0);

	link_real(0, "pmark", &pmark_, 1);

	// Add read parameter pmark_
}

void BlueQueue::reset()
{
	Link::reset();
	/*
	 * Packet time constant -- units are pkts/sec.
	 */
	ptc_ = ( bandwidth_ / ( 8. * mean_pktsize_ ));

	idle_ = 1;
	idletime_ = Scheduler::instance().clock();

	pmark_ = 0;
	ifreezetime_ = 0;
	dfreezetime_ = 0;

}

/*
 * Output the average queue size and the dropping probability. 
 */
void BlueQueue::plot()
{
	double t = Scheduler::instance().clock();
	// sprintf(trace_->buffer(), "a %g %g", t, edv_.v_ave);
	// trace_->dump();
	// sprintf(trace_->buffer(), "p %g %g", t, edv_.v_prob);
	// trace_->dump();
}

/*
 * print the queue seen by arriving packets only
 */
void BlueQueue::plot1(int length)
{
        double t =  Scheduler::instance().clock();
        sprintf(trace_->buffer(), "Q %g %d", t, length);
	trace_->dump();
}

/*
 * print pmark when updated
 */
void BlueQueue::pmark_plot(int method)
{
        double t =  Scheduler::instance().clock();
        if (method)
            sprintf(trace_->buffer(), "I %g %g", t, pmark_);
        else
            sprintf(trace_->buffer(), "D %g %g", t, pmark_);
	trace_->dump();
}

void BlueQueue::decrement_pmark(int how)
{
    double now = Scheduler::instance().clock();

    if (now - dfreezetime_ > dholdtime_) {
    	dfreezetime_ = now;
    	switch (dalgorithm_) {
    	case 0:
	case 2:
		switch (how) {
			case 0:
        			pmark_ -= decrement_;
				break;
			case 1:
			default:
				break;
		}
		break;
    	default:
    	case 1:
		switch (how) {
			case 0:
        		pmark_ -= decrement_;
				break;
			case 1:
			default:
        			pmark_ -= decrement_/10;
				break;
		}
    	}
    	if (pmark_ < 0)
		pmark_ = 0.0;
    }
}

void BlueQueue::increment_pmark(int how)
{
    double now = Scheduler::instance().clock();

    if (now - ifreezetime_ > iholdtime_) {
    	ifreezetime_ = now;
    	switch (ialgorithm_) {
    	case 0:
		switch (how) {
			case 0:
        			pmark_ += increment_;
				break;
			case 1:
			default:
				break;
		}
		break;
	case 2:
		switch (how) {
			case 0:
        			pmark_ = 2*pmark_ + increment_;
				break;
			case 1:
			default:
				break;
		}
    	default:
    	case 1:
		switch (how) {
			case 0:
        			pmark_ += increment_;
				break;
			case 1:
			default:
        			pmark_ += increment_/10;
				break;
		}
    	}
    	if (pmark_ > 1.0) 
		pmark_ = 1.00;
    }
}

/*
 * Return the next packet in the queue for transmission.
 */
Packet* BlueQueue::deque()
{
	Packet* p = q_.deque();
	double now = Scheduler::instance().clock();

	if (p != 0) {
		idle_ = 0;
		log_packet_departure(p);
	} else {
		decrement_pmark(0);
        	if (trace_)
        		pmark_plot(0);
		idle_ = 1;
		idletime_ = Scheduler::instance().clock();
	}
	return (p);
}
		
int BlueQueue::drop_early(Packet* pkt)
{
	double u = Random::uniform();
	if (u <= pmark_) {
		if (setbit_) {
			pkt->flags_ |= PF_ECN;
		}
		else
			return (1);
	}
	return (0);
}

/*
 * Receive a new packet arriving at the queue.
 * The average queue size is computed.  If the average size
 * exceeds the threshold, then the dropping probability is computed,
 * and the newly-arriving packet is dropped with that probability.
 * The packet is also dropped if the maximum queue size is exceeded.
 */
void BlueQueue::enque(Packet* pkt)
{
	double now = Scheduler::instance().clock();
	int m;
	Flow *f = 0;

        if (idle_) {
		/* To account for the period when the queue was empty.  */
                idle_ = 0;
		m = int(ptc_ * (now - idletime_));
        } else
                m = 0;

	if (trace_)
		if (bytes_) {
			plot1(qnb_);
		}
		else {
			plot1(qnp_);
		}

	if (flowmgr_) {
		if ((f = flowmgr_->findflow(pkt)) == NULL)
			f = flowmgr_->newflow(pkt);
		if (f)
			f->arrival(pkt);
	}

        /*
         * Drop each packet with probability edv.v_prob.
         */
        if (drop_early(pkt)) {
	    double now = Scheduler::instance().clock();
            log_packet_arrival(pkt);
            log_packet_drop(pkt);
            f->drop(pkt,1);
            // if (f) {
            //    if (edv_.v_ave < edp_.th_max)
            //        f->drop(pkt, 1);
            //    else f->drop(pkt, 0);
            // }
	    Packet::free(pkt);
            pkt = 0;
        }
	/*
	 * If we didn't drop the packet above, send it to the interface,
	 * checking for absolute queue overflow.
	 */
	if (pkt != 0) {
		q_.enque(pkt);
		log_packet_arrival(pkt);
		int metric = bytes_ ? qnb_ : qnp_;
		int limit = bytes_ ? (qlim_* mean_pktsize_): qlim_;
		int half = limit/2;
		// if (metric > half)
			// increment_pmark(1);	
		// else
			// decrement_pmark(1);
		if (metric > limit) {
	    		double now = Scheduler::instance().clock();
			pkt = q_.lookup(qnp_ - 1);
			q_.remove(pkt);
			increment_pmark(0);
                        if (trace_)
                        	pmark_plot(1);
			log_packet_drop(pkt);
			if (f)
				f->drop(pkt, 0);
			Packet::free(pkt);
		}
	}
	if (trace_)
		plot();
}

class PBlueQueue : public Link {
 public:	
	PBlueQueue();
	inline int qlen() { return(q_.length()); }

 protected:
	virtual void reset();
	virtual void plot();
	void plot1(int qlen);
	void pmark_plot(int method);
	virtual int drop_early(Packet* pkt);
	virtual void enque(Packet* pkt);
	virtual Packet* deque();
	void increment_pmark(int how);
	void decrement_pmark(int how);
	void scale_pmark();
	void reset_counts();
	Queue q_;
	/*
	 * Static state.
	 */
	int bytes_;	/* bytes or packets? */
	int dummy_;	/* */
	int setbit_;
	int mean_pktsize_;
	double decrement_;
	double increment_;
	double iholdtime_;
	double dholdtime_;
	bool flowarray[MAXFLOWS];

	/*
	 * Dynamic state.
	 */
	int idle_;
	double idletime_;
	double ptc_;	/* packet time constant in packets/second */
	double ifreezetime_;
	double dfreezetime_;
	double pmark_;
};

static class PBlueMatcher : public Matcher {
public:
	PBlueMatcher() : Matcher("link") {}
	TclObject* match(const char* id) {
		if (strcasecmp(id, "pblue") == 0)
			return (new PBlueQueue);
		return (0);
	}
} matcher_blue;

PBlueQueue::PBlueQueue()
{
	Tcl& tcl = Tcl::instance();

	link_bool("ns_blue", "bytes", &bytes_ , 0);
	link_bool("ns_blue", "setbit", &setbit_ , 0);
	link_real("ns_blue", "decrement", &decrement_ , 0);
	link_real("ns_blue", "increment", &increment_ , 0);
	link_time("ns_blue", "dhold-time", &dholdtime_ , 0);
	link_time("ns_blue", "ihold-time", &iholdtime_ , 0);

	link_real(0, "pmark", &pmark_, 1);

	// Add read parameter pmark_
}

void PBlueQueue::reset()
{
	Link::reset();
	/*
	 * Packet time constant -- units are pkts/sec.
	 */
	ptc_ = ( bandwidth_ / ( 8. * mean_pktsize_ ));

	idle_ = 1;
	idletime_ = Scheduler::instance().clock();

	pmark_ = 0;
	ifreezetime_ = 0;
	dfreezetime_ = 0;

}

/*
 * Output the average queue size and the dropping probability. 
 */
void PBlueQueue::plot()
{
	double t = Scheduler::instance().clock();
	// sprintf(trace_->buffer(), "a %g %g", t, edv_.v_ave);
	// trace_->dump();
	// sprintf(trace_->buffer(), "p %g %g", t, edv_.v_prob);
	// trace_->dump();
}

/*
 * print the queue seen by arriving packets only
 */
void PBlueQueue::plot1(int length)
{
        double t =  Scheduler::instance().clock();
        sprintf(trace_->buffer(), "Q %g %d", t, length);
	trace_->dump();
}

/*
 * print pmark when updated
 */
void PBlueQueue::pmark_plot(int method)
{
        double t =  Scheduler::instance().clock();
        if (method)
            sprintf(trace_->buffer(), "I %g %g", t, pmark_);
        else
            sprintf(trace_->buffer(), "D %g %g", t, pmark_);
	trace_->dump();
}

void PBlueQueue::decrement_pmark(int how)
{
    double now = Scheduler::instance().clock();

    if (now - dfreezetime_ > dholdtime_) {
    	dfreezetime_ = now;
        pmark_ -= decrement_;
    	if (pmark_ < 0)
		pmark_ = 0.0;
    }
}

void PBlueQueue::increment_pmark(int how)
{
    double now = Scheduler::instance().clock();

    if (now - ifreezetime_ > iholdtime_) {
    	ifreezetime_ = now;
        pmark_ += increment_;
    	if (pmark_ > 1.0) 
		pmark_ = 1.00;
    }
}

/*
 * Return the next packet in the queue for transmission.
 */
Packet* PBlueQueue::deque()
{
	Packet* p = q_.deque();
	double now = Scheduler::instance().clock();

	if (p != 0) {
		idle_ = 0;
		log_packet_departure(p);
	} else {
		decrement_pmark(0);
        	if (trace_)
        		pmark_plot(0);
		idle_ = 1;
		idletime_ = Scheduler::instance().clock();
	}
	return (p);
}
		
int PBlueQueue::drop_early(Packet* pkt)
{
	double u = Random::uniform();
	if (u <= pmark_) {
		if (setbit_) {
			pkt->flags_ |= PF_ECN;
		}
		else
			return (1);
	}
	return (0);
}

/*
 * Receive a new packet arriving at the queue.
 * The average queue size is computed.  If the average size
 * exceeds the threshold, then the dropping probability is computed,
 * and the newly-arriving packet is dropped with that probability.
 * The packet is also dropped if the maximum queue size is exceeded.
 */
void PBlueQueue::enque(Packet* pkt)
{
	double now = Scheduler::instance().clock();
	int m;
	Flow *f = 0;

        if (idle_) {
		/* To account for the period when the queue was empty.  */
                idle_ = 0;
		m = int(ptc_ * (now - idletime_));
        } else
                m = 0;

	if (trace_)
		if (bytes_) {
			plot1(qnb_);
		}
		else {
			plot1(qnp_);
		}

	if (flowmgr_) {
		if ((f = flowmgr_->findflow(pkt)) == NULL)
			f = flowmgr_->newflow(pkt);
		if (f)
			f->arrival(pkt);
	}

        /*
         * Drop each packet with probability edv.v_prob.
         */
        if (drop_early(pkt)) {
	    double now = Scheduler::instance().clock();
            log_packet_arrival(pkt);
            log_packet_drop(pkt);
            f->drop(pkt,1);
            // if (f) {
            //    if (edv_.v_ave < edp_.th_max)
            //        f->drop(pkt, 1);
            //    else f->drop(pkt, 0);
            // }
	    Packet::free(pkt);
            pkt = 0;
        }
	/*
	 * If we didn't drop the packet above, send it to the interface,
	 * checking for absolute queue overflow.
	 */
	if (pkt != 0) {
		q_.enque(pkt);
		log_packet_arrival(pkt);
		int metric = bytes_ ? qnb_ : qnp_;
		int limit = bytes_ ? (qlim_* mean_pktsize_): qlim_;
		int half = limit/2;
		// if (metric > half)
			// increment_pmark(1);	
		// else
			// decrement_pmark(1);
		if (metric > limit) {
	    		double now = Scheduler::instance().clock();
			pkt = q_.lookup(qnp_ - 1);
			q_.remove(pkt);
			increment_pmark(0);
                        if (trace_)
                        	pmark_plot(1);
			log_packet_drop(pkt);
			if (f)
				f->drop(pkt, 0);
			Packet::free(pkt);
		}
	}
	if (trace_)
		plot();
}
