/*
 * Copyright (c) 1990-1994 Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by the Computer Systems
 *	Engineering Group at Lawrence Berkeley Laboratory.
 * 4. Neither the name of the University nor of the Laboratory may be used
 *    to endorse or promote products derived from this software without
 *    specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */

#ifndef lint
static char rcsid[] =
    "@(#) $Header: /root/ns-1.1/RCS/fairblue.cc,v 1.1 1999/01/08 20:27:45 root Exp $ (LBL)";
#endif

#include <math.h>
#include <string.h>
#include <sys/types.h>

#include "link.h"
#include "Tcl.h"
#include "packet.h"
#include "trace.h"
#include "flow.h"
#include "random.h"
#include "scheduler.h"

extern "C" { int strcasecmp(const char*, const char*); }

#define SFQ_BINS 23
#define MIN(a,b)	((a) > (b) ? (b) : (a))
#define MAX(a,b)	((a) > (b) ? (a) : (b))
#define TCP_RTO 0.200
#define PM_TH 0.98
#define UNPENALIZE 0.10

unsigned int hashit(unsigned int a, unsigned int b, unsigned int c, unsigned int d, int modulus)
{
        unsigned int h;
        h = ((a << 6) + (b << 12) + (c << 18) + (d << 24) ) % modulus;
        return h;
}

int hash0(Packet* pkt, unsigned int fudge)
{
	unsigned int i,j;
	j = (unsigned int) pkt->dport_ * fudge;
	i = hashit(pkt->dst_, pkt->sport_, pkt->src_, 
			j, SFQ_BINS);
	return(i);
}

int hash1(Packet* pkt, unsigned int fudge)
{
	unsigned int i,j;
	j = (unsigned int) pkt->sport_ * fudge;
	i = hashit(pkt->src_, pkt->dport_, j,
			pkt->dst_, SFQ_BINS);
	return(i);
}

struct sfqbin {
	double	pmark;
	double  freezetime;
	double	penalizetime;
	int	pkts;
};

class FairBlueQueue : public Link {
 public:	
	FairBlueQueue();
	inline int qlen() { return(q_.length()); }

 protected:
	virtual void reset();
	virtual void plot();
	virtual void mark_ecn(Packet* pkt);
	virtual int drop_early(Packet* pkt);
	virtual void enque(Packet* pkt);
	virtual Packet* deque();
	void plot1(Packet* pkt);
	void reset_bins();
	void eq_update_bins(Packet* pkt);
	void dq_update_bins(Packet* pkt);
	void decrement_bins(Packet* pkt, double decrement, int force);
	void decrement_bin(Packet* pkt, double decrement, int force, int i);
	void increment_bins(Packet* pkt, double increment, int force);
	void increment_bin(Packet* pkt, double increment, int force, int i);
	int penalize(Packet* pkt);
	int pcheck(Packet* pkt);
	Queue q_;
	/*
	 * Static state.
	 */
	int bytes_;	/* bytes or packets? */
	int mean_pktsize_;
	double decrement_;
	double increment_;
	double holdtime_;
	double pboxtime_;
	int hinterval_;
	double pboxfreeze_;
	unsigned int clearpkts_;
	int algorithm_;
	int pbox_;
	int allocation_;
	int drop_thresh_;
	unsigned int last_fudge_;

	/*
	 * Dynamic state.
	 */
	int idle_;
	double idletime_;
	double ptc_;	/* packet time constant in seconds/packet */
	struct sfqbin bins[SFQ_LEVELS][SFQ_BINS];
	int pbox_count_;
};

static class FairBlueMatcher : public Matcher {
public:
	FairBlueMatcher() : Matcher("link") {}
	TclObject* match(const char* id) {
		if (strcasecmp(id, "fairblue") == 0)
			return (new FairBlueQueue);
		return (0);
	}
} matcher_fairblue;

FairBlueQueue::FairBlueQueue()
{
	Tcl& tcl = Tcl::instance();

	link_bool("ns_fairblue", "bytes", &bytes_ , 0);
	link_real("ns_fairblue", "decrement", &decrement_ , 0);
	link_real("ns_fairblue", "increment", &increment_ , 0);
	link_time("ns_fairblue", "hold-time", &holdtime_ , 0);
	link_time("ns_fairblue", "pbox-time", &pboxtime_ , 0);

	link_int("ns_fairblue", "hinterval", &hinterval_ , 0);
	link_int("ns_fairblue", "algorithm", &algorithm_ , 0);
	link_int("ns_fairblue", "pbox", &pbox_ , 0);
}

void FairBlueQueue::reset()
{
	Link::reset();
        double t =  Scheduler::instance().clock();

	/*
	 * Packet time constant -- units are secs/pkt.
	 */
	ptc_ = 8000. / bandwidth_; 

	idle_ = 1;
	idletime_ = Scheduler::instance().clock();

	for (int i=0; i < SFQ_BINS; i++)
		for (int j=0; j < SFQ_LEVELS; j++) {
			bins[j][i].pkts = 0;
			bins[j][i].pmark = 0;
			bins[j][i].freezetime = 0;
			bins[j][i].penalizetime = 0;
		}
	allocation_ = qlim_/SFQ_BINS;
	drop_thresh_ = allocation_ + allocation_/2;
	if (trace_) {
		sprintf(trace_->buffer(),"alloc[%d] dthresh[%d] ptc[%g] qlim[%d]", 
			allocation_, drop_thresh_, ptc_, qlim_);
		trace_->dump();
	}

	clearpkts_ = 0;
	pbox_count_ = 0;
	pboxfreeze_ = 0;
	last_fudge_ = 1;
}

void FairBlueQueue::plot()
{
}

void FairBlueQueue::reset_bins()
{
	clearpkts_ = qlen();
	for (int i=0; i < SFQ_BINS; i++)
		for (int j=0; j < SFQ_LEVELS; j++) {
			bins[j][i].pkts = 0;
			bins[j][i].freezetime = 0;
			bins[j][i].pmark = 0;
		}
}

/*
 * Output the average queue size and the dropping probability. 
 */
void FairBlueQueue::plot1(Packet* p)
{
	double t = Scheduler::instance().clock();
	sprintf(trace_->buffer(),"  [%d %d %d %d] -> [%d %d] %g %g",
	        p->src_, p->sport_, p->dst_, p->dport_,
		p->h[0], p->h[1],
                bins[0][p->h[0]].pmark,
                bins[1][p->h[1]].pmark);
	trace_->dump();
}

void FairBlueQueue::dq_update_bins(Packet* pkt)
{
   double now;
   for (int i=0; i < SFQ_LEVELS; i++) {
	bins[i][pkt->h[i]].pkts--;
	if (bins[i][pkt->h[i]].pkts <= 0) {
	   now = Scheduler::instance().clock();
	   bins[i][pkt->h[i]].pkts = 0;
	   decrement_bin(pkt,decrement_, 0, i);
	}
   }
}

void FairBlueQueue::eq_update_bins(Packet* pkt)
{
	for (int i=0; i < SFQ_LEVELS; i++)
		bins[i][pkt->h[i]].pkts++;
}

void FairBlueQueue::increment_bin(Packet* pkt, double increment, int force, int i)
{
	double now = Scheduler::instance().clock();
	switch (algorithm_) {
	   case 1:
	   case 2:
	   case 3:
	  	if ((now - bins[i][pkt->h[i]].freezetime < holdtime_) && !force)
		   break;
		bins[i][pkt->h[i]].freezetime = now;
	   case 0:
	   default:
		bins[i][pkt->h[i]].pmark += increment;
		if (bins[i][pkt->h[i]].pmark > 1.00)
			bins[i][pkt->h[i]].pmark = 1.00;
	}
}

void FairBlueQueue::increment_bins(Packet* pkt, double increment, int force)
{
	double now = Scheduler::instance().clock();
	for (int i=0; i < SFQ_LEVELS; i++) {
	   switch (algorithm_) {
		case 1:
		case 2:
	   	case 3:
	   		if ((now - bins[i][pkt->h[i]].freezetime < holdtime_) && !force)
			   break;
			bins[i][pkt->h[i]].freezetime = now;
		case 0:
		default:
			bins[i][pkt->h[i]].pmark += increment;
			if (bins[i][pkt->h[i]].pmark > 1.00)
				bins[i][pkt->h[i]].pmark = 1.00;
	   }
	}
}

void FairBlueQueue::decrement_bin(Packet* pkt, double decrement, int force, int i)
{
	double now = Scheduler::instance().clock();
	switch (algorithm_) {
	   case 1:
	   case 2:
	   case 3:
		if ((now - bins[i][pkt->h[i]].freezetime < holdtime_) && !force)
		   break;
		bins[i][pkt->h[i]].freezetime = now;
	   case 0:
	   default:
		bins[i][pkt->h[i]].pmark -= decrement;
		if (bins[i][pkt->h[i]].pmark < 0)
			bins[i][pkt->h[i]].pmark = 0;
	}
}

void FairBlueQueue::decrement_bins(Packet* pkt, double decrement, int force)
{
	double now = Scheduler::instance().clock();
	for (int i=0; i < SFQ_LEVELS; i++) {
	   switch (algorithm_) {
		case 1:
		case 2:
	   	case 3:
	   		if ((now - bins[i][pkt->h[i]].freezetime < holdtime_) && !force)
			   break;
			bins[i][pkt->h[i]].freezetime = now;
		case 0:
		default:
			bins[i][pkt->h[i]].pmark -= decrement_;
			if (bins[i][pkt->h[i]].pmark < 0)
				bins[i][pkt->h[i]].pmark = 0;
	   }
	}
}

/*
 * Return the next packet in the queue for transmission.
 */
Packet* FairBlueQueue::deque()
{
	Packet* pkt = q_.deque();
	double now = Scheduler::instance().clock();

	if (pkt != 0) {
		if (pkt->pbox_) {
		  pkt->pbox_ = 0;
		  pbox_count_--;
		  if (pbox_count_ < 0) pbox_count_ = 0;
		  if (clearpkts_ > 0) clearpkts_--;
		}
		else {
		  	if (clearpkts_ > 0) clearpkts_--;
			else dq_update_bins(pkt);
		}
		idle_ = 0;
		log_packet_departure(pkt);
	} else {
		idle_ = 1;
		idletime_ = Scheduler::instance().clock();
	}
	return (pkt);
}
		
int FairBlueQueue::drop_early(Packet* pkt)
{
   int return_code;
   double now = Scheduler::instance().clock();

   return_code = 0;
	
   for (int i=0; i < SFQ_LEVELS; i++) {
	if (bins[i][pkt->h[i]].pkts >= allocation_) {
	   if (bins[i][pkt->h[i]].pkts >= drop_thresh_)
		return_code = 1;
	   increment_bin(pkt, increment_, 0, i);
	   if (trace_) {
		// sprintf(trace_->buffer(), 
                //       "level %d over(%g) <%d> [%d %d %d %d] -> [%d %d] %g %g",
                // 	i, now, qlen(), pkt->src_, pkt->sport_, pkt->dst_, 
		//	pkt->dport_, pkt->h[0], pkt->h[1], 
		//	bins[0][pkt->h[0]].pmark, bins[1][pkt->h[1]].pmark);
		// trace_->dump();
	   }
	}
   }
   return(return_code);
}

void FairBlueQueue::mark_ecn(Packet* pkt)
{
	double u = Random::uniform();
	double p;
	double now = Scheduler::instance().clock();

	p = MIN(bins[0][pkt->h[0]].pmark, bins[1][pkt->h[1]].pmark);

	if (u <= p) {
		pkt->flags_ |= PF_ECN;
		if (trace_) {
			sprintf(trace_->buffer(), 
			"ecn mark(%g) <%d> [%d %d %d %d] -> [%d %d] %g %g",
                		now, qlen(), pkt->src_, pkt->sport_, pkt->dst_, 
				pkt->dport_, pkt->h[0], pkt->h[1], 
				bins[0][pkt->h[0]].pmark,
                		bins[1][pkt->h[1]].pmark);
			trace_->dump();
		}
	}

}

int FairBlueQueue::pcheck(Packet* pkt)
{
    // Returns 1 if all pmark > 0.98 and all bins are not empty
    // Returns 0 otherwise
    // if ((bins[i][pkt->h[i]].pmark < PM_TH) || (bins[i][pkt->h[i]].pkts == 0))
    for (int i=0; i < SFQ_LEVELS; i++) {
        if (bins[i][pkt->h[i]].pmark < PM_TH)
            return(0);
    }
    return(1);
}

int FairBlueQueue::penalize(Packet* pkt)
{
  double now =  Scheduler::instance().clock();
  double last_penalty, p;
  double scale;
  int low;

  if (pcheck(pkt)) {
	if (bins[0][pkt->h[0]].penalizetime < bins[1][pkt->h[1]].penalizetime)
	   low = 0;
	else
	   low = 1;
	last_penalty = bins[low][pkt->h[low]].penalizetime;
	if ((now - last_penalty) > TCP_RTO) {
            decrement_bin(pkt,UNPENALIZE,1,low);
            if (trace_) {
		sprintf(trace_->buffer(), 
                "unpenalize (%g) <%d> [%d %d %d %d] -> [%d %d] %g %g %g", 
                        now, qlen(), pkt->src_, pkt->sport_, pkt->dst_, 
			pkt->dport_, pkt->h[0], pkt->h[1],
			bins[0][pkt->h[0]].pmark, bins[1][pkt->h[1]].pmark,
			scale);
		trace_->dump();
	    }
            bins[0][pkt->h[0]].penalizetime = now;
            bins[1][pkt->h[1]].penalizetime = now;
	    return(0);
	}
	switch (algorithm_) {
	   case 2:
	   case 3:
		if ((now - pboxfreeze_) < pboxtime_) {
	   		increment_bins(pkt,increment_,1);
	   		bins[0][pkt->h[0]].penalizetime = now;
	   		bins[1][pkt->h[1]].penalizetime = now;
			return(1);
		}
            	else {
	   		pkt->pbox_ = 1;
			pboxfreeze_ = now;
			return(0);
		}
		break;
	   case 0:
	   case 1:
	   default:
		if (pbox_count_ < pbox_) {
		   pkt->pbox_ = 1;
		   return(0);
		} else {
		   increment_bins(pkt,increment_,1);
		   bins[0][pkt->h[0]].penalizetime = now;
		   bins[1][pkt->h[1]].penalizetime = now;
		   return(1);
		}
		break;
  	}
  }
  return(0);
}

/*
 * Receive a new packet arriving at the queue.
 * The average queue size is computed.  If the average size
 * exceeds the threshold, then the dropping probability is computed,
 * and the newly-arriving packet is dropped with that probability.
 * The packet is also dropped if the maximum queue size is exceeded.
 */
void FairBlueQueue::enque(Packet* pkt)
{
	double now = Scheduler::instance().clock();
	int m;
	unsigned int fudge;
	Flow *f = 0;

	switch (algorithm_) {
		case 3:
			fudge = (unsigned int) now/hinterval_;
			if (fudge != last_fudge_) {
				last_fudge_ = fudge;
				reset_bins();
			}
			break;
		case 0:
		case 1:
		case 2:
		default:
			fudge = 1;
			last_fudge_ = 1;
			break;
	}
	pkt->h[0] = hash0(pkt,fudge);
	pkt->h[1] = hash1(pkt,fudge);

	if (flowmgr_) {
		if ((f = flowmgr_->findflow(pkt)) == NULL)
			f = flowmgr_->newflow(pkt);
		if (f)
			f->arrival(pkt);
	}

	if (penalize(pkt)) {
	    double now = Scheduler::instance().clock();
            log_packet_arrival(pkt);
            log_packet_drop(pkt);
            if (f) f->drop(pkt,1);
	    if (trace_) {
		sprintf(trace_->buffer(), 
                       "penalize drop(%g) <%d> [%d %d %d %d] -> [%d %d] %g %g", 
                        now, qlen(), pkt->src_, pkt->sport_, pkt->dst_, 
			pkt->dport_, pkt->h[0], pkt->h[1],
			bins[0][pkt->h[0]].pmark, bins[1][pkt->h[1]].pmark);
		trace_->dump();
	    }
	    Packet::free(pkt);
	    pkt = 0;
	}
	else {
        	/*
         	 * Mark each packet with probability edv.v_prob.
         	 */
		if (pkt->pbox_ == 0) {
        		mark_ecn(pkt);
			if (drop_early(pkt)) {
	    			double now = Scheduler::instance().clock();
            			log_packet_arrival(pkt);
            			log_packet_drop(pkt);
            			if (f) f->drop(pkt,0);
	    			pkt = 0;
			}
		}
        }
	/*
	 * If we didn't drop the packet above, send it to the interface,
	 * checking for absolute queue overflow.
	 */
	if (pkt != 0) {
		q_.enque(pkt);
		log_packet_arrival(pkt);
		int metric = bytes_ ? qnb_ : qnp_;
		int limit = bytes_ ? (qlim_* mean_pktsize_): qlim_;
		if (metric > limit) {
	    		double now = Scheduler::instance().clock();
			if (trace_) {
			   sprintf(trace_->buffer(), 
			   "fullq drop(%g) <%d> [%d %d %d %d] -> [%d %d] %g %g",
				now, qlen(), pkt->src_, pkt->sport_, pkt->dst_, 
				pkt->dport_,pkt->h[0], pkt->h[1],
				bins[0][pkt->h[0]].pmark,
				bins[1][pkt->h[1]].pmark);
			   trace_->dump() ;
			}
			pkt = q_.lookup(qnp_ - 1);
			q_.remove(pkt);
			log_packet_drop(pkt);
			if (f)
				f->drop(pkt, 0);
			Packet::free(pkt);
			pkt = 0;
		} else {
			if (pkt->pbox_ == 0) 
				eq_update_bins(pkt);
			else
				pbox_count_++;
		}
	}
}
