/*
 * Copyright (c) 1990-1994 Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by the Computer Systems
 *	Engineering Group at Lawrence Berkeley Laboratory.
 * 4. Neither the name of the University nor of the Laboratory may be used
 *    to endorse or promote products derived from this software without
 *    specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *
 * Here is one set of parameters from one of my simulations:
 * 
 * ed [ q_weight=0.002 thresh=5 linterm=30 maxthresh=15
 *         mean_pktsize=500 dropmech=random-drop queue-size=60
 *         plot-file=none bytes=false doubleq=false dqthresh=50 
 *	   wait=true ]
 * 
 * 1/"linterm" is the max probability of dropping a packet. 
 * There are different options that make the code
 * more messy that it would otherwise be.  For example,
 * "doubleq" and "dqthresh" are for a queue that gives priority to
 *   small (control) packets, 
 * "bytes" indicates whether the queue should be measured in bytes 
 *   or in packets, 
 * "dropmech" indicates whether the drop function should be random-drop 
 *   or drop-tail when/if the queue overflows, and 
 *   the commented-out Holt-Winters method for computing the average queue 
 *   size can be ignored.
 * "wait" indicates whether the gateway should wait between dropping
 *   packets.
 */

#ifndef lint
static char rcsid[] =
    "@(#) $Header: /n/olympus/opt/users/CVS/ns-1.1-master/red.cc,v 1.5 1998/09/09 19:10:34 wuchang Exp $ (LBL)";
#endif

#include <math.h>
#include <string.h>
#include <sys/types.h>

#include "link.h"
#include "Tcl.h"
#include "packet.h"
#include "trace.h"
#include "flow.h"
#include "random.h"
#include "scheduler.h"

#define MAXCLASS	256
#define MAXQ		16

extern "C" { int strcasecmp(const char*, const char*); }

/*
 * Early drop parameters.
 */
struct edp {
	/*
	 * User supplied.
	 */
	int mean_pktsize;
	int bytes;		/* true if queue in bytes, false if packets */
	int wait;		/* true for waiting between dropped packets */
	int setbit;		/* true to set congestion indication bit */

	double th_min;		/* minimum threshold */
	double th_max;		/* maximum threshold */
	double max_p_inv;	/* 1/max_p, for max_p = maximum prob.  */
	double q_w;		/* queue weight */
	/*
	 * Computed as a function of user supplied paramters.
	 */
	double ptc;		/* packet time constant in packets/second */
};

/*
 * Early drop variables.
 */
struct edv {
	float v_ave;		/* average queue size */
	float v_slope;		/* used in computing average queue size */
	float v_r;			
	float v_prob;		/* prob. of packet drop */
	struct dlist* drops;
	float v_a;		/* v_prob = v_a * v_ave + v_b */
	float v_b;
	int count;		/* # of packets since last drop */
	int count_bytes;	/* # of bytes since last drop */
	int old;		/* 0 when average queue first exceeds thresh */
	double max_p_inv;	/* 1/max_p, for max_p = maximum prob.  */
	int status;		/* 0=below min_th, 1=between, 2=above max_th */
	int pkts;		/* For SFRED */
	int idle;		/* For SFRED */
	double idletime;
};

struct q_stat {
	int valid;
	u_short src;
	u_short dst;
	u_short sport;
	u_short dport;
	int unmarked;
	int marked;
	int rlimit;
	int strikes;
};

class REDQueue : public Link {
 public:	
	REDQueue();
	inline int qlen() { return(q_.length()); }

 protected:
	virtual void reset();
	virtual void run_estimator(int nqueued, int m);
	virtual void plot();
	void plot1(int qlen);
	virtual int drop_early(Packet* pkt);
	virtual void enque(Packet* pkt);
	virtual Packet* deque();
	Queue q_;
	/*
	 * Static state.
	 */
	int drop_tail_;
	edp edp_;
	int doubleq_;	/* for experiments with priority for small packets */
	int dqthresh_;	/* for experiments with priority for small packets */

	/*
	 * Dynamic state.
	 */
	int idle_;
	double idletime_;
	edv edv_;
};

static class REDMatcher : public Matcher {
public:
	REDMatcher() : Matcher("link") {}
	TclObject* match(const char* id) {
		if (strcasecmp(id, "red") == 0)
			return (new REDQueue);
		return (0);
	}
} matcher_red;

REDQueue::REDQueue()
{
	Tcl& tcl = Tcl::instance();

	link_bool("ns_red", "bytes", &edp_.bytes, 0);
	link_real("ns_red", "thresh", &edp_.th_min, 0);
	link_real("ns_red", "maxthresh", &edp_.th_max, 0);
	link_int("ns_red", "mean_pktsize", &edp_.mean_pktsize, 0);
	link_real("ns_red", "q_weight", &edp_.q_w, 0);
	link_bool("ns_red", "wait", &edp_.wait, 0);
	link_real("ns_red", "linterm", &edp_.max_p_inv, 0);
	link_bool("ns_red", "setbit", &edp_.setbit, 0);
	link_bool("ns_red", "drop-tail", &drop_tail_, 0);

	link_bool("ns_red", "doubleq", &doubleq_, 0);
	link_int("ns_red", "dqthresh", &dqthresh_, 0);
}

void REDQueue::reset()
{
	Link::reset();
	/*
	 * Packet time constant -- units are pkts/sec.
	 */
	edp_.ptc = ( bandwidth_ / ( 8. * edp_.mean_pktsize ));

        if (edp_.bytes) {
		edp_.th_min *= edp_.mean_pktsize;
		edp_.th_max *= edp_.mean_pktsize;
	}

	edv_.v_ave = 0.0;
	edv_.v_slope = 0.0;
	edv_.drops = 0;
	edv_.count = 0;
	edv_.count_bytes = 0;
	edv_.old = 0;
	edv_.v_a = 1 / (edp_.th_max - edp_.th_min);
	edv_.v_b = - edp_.th_min / (edp_.th_max - edp_.th_min);

	idle_ = 1;
	idletime_ = Scheduler::instance().clock();

}

/*
 * Compute the average queue size.
 * The code contains two alternate methods for this, the plain EWMA
 * and the Holt-Winters method.
 * nqueued can be bytes or packets
 */
void REDQueue::run_estimator(int nqueued, int m)
{
	float f, f_sl, f_old;

	f = edv_.v_ave;
	f_sl = edv_.v_slope;
#define RED_EWMA
#ifdef RED_EWMA
        while (--m >= 1) {
                f_old = f;
                f *= 1.0 - edp_.q_w;
        }
        f_old = f;
        f *= 1.0 - edp_.q_w;
        f += edp_.q_w * nqueued;
#endif
#ifdef RED_HOLT_WINTERS
	while (--m >= 1) {
		f_old = f;
		f += f_sl;
		f *= 1.0 - edp_.q_w;
		f_sl *= 1.0 - 0.5 * edp_.q_w;
		f_sl += 0.5 * edp_.q_w * (f - f_old);
	}
	f_old = f;
	f += f_sl;
	f *= 1.0 - edp_.q_w;
	f += edp_.q_w * nqueued;
	f_sl *= 1.0 - 0.5 * edp_.q_w;
	f_sl += 0.5 * edp_.q_w * (f - f_old);
#endif
	edv_.v_ave = f;
	edv_.v_slope = f_sl;
}

/*
 * Output the average queue size and the dropping probability. 
 */
void REDQueue::plot()
{
	double t = Scheduler::instance().clock();
	sprintf(trace_->buffer(), "a %g %g", t, edv_.v_ave);
	trace_->dump();
	sprintf(trace_->buffer(), "p %g %g", t, edv_.v_prob);
	trace_->dump();
}

/*
 * print the queue seen by arriving packets only
 */
void REDQueue::plot1(int length)
{
        double t =  Scheduler::instance().clock();
        sprintf(trace_->buffer(), "Q %g %d", t, length);
	trace_->dump();
}

/*
 * Return the next packet in the queue for transmission.
 */
Packet* REDQueue::deque()
{
	Packet* p = q_.deque();
	if (p != 0) {
		idle_ = 0;
		log_packet_departure(p);
	} else {
		idle_ = 1;
		idletime_ = Scheduler::instance().clock();
	}
	return (p);
}
		
int REDQueue::drop_early(Packet* pkt)
{
	if (edv_.v_ave >= edp_.th_max)
		edv_.v_prob = 1.0;
	else {
		double p = edv_.v_a * edv_.v_ave + edv_.v_b;
		double count1 = edv_.count;
		if ( edp_.bytes )
			count1 = (double) (edv_.count_bytes/edp_.mean_pktsize);
		p /= edp_.max_p_inv;
		if (edp_.wait ) {
			if (count1 * p < 1)
				p = 0;
			else if (count1 * p < 2)
				p /= (2 - count1 * p);
			else
				p = 1;
		} else if (!edp_.wait ) {
			if (count1 * p < 1)
				p /= (1 - count1 * p);
			else
				p = 1;
		}
		if (edp_.bytes && p < 1 ) {
			p = p * pkt->size_ / edp_.mean_pktsize;
		}
		if (p>1) p=1.0;
		edv_.v_prob = p;
#ifdef DEBUGBYTES
		if (p > 0)
			printf ("bytes: %6d prob: %g\n", pkt->size_, p);
#endif
	}
	double u = Random::uniform();
	if (u <= edv_.v_prob) {
		edv_.count = 0;
		edv_.count_bytes = 0;
		if (edp_.setbit) 
			pkt->flags_ |= PF_ECN;
		else
			return (1);
	}
	return (0);
}

/*
 * Receive a new packet arriving at the queue.
 * The average queue size is computed.  If the average size
 * exceeds the threshold, then the dropping probability is computed,
 * and the newly-arriving packet is dropped with that probability.
 * The packet is also dropped if the maximum queue size is exceeded.
 */
void REDQueue::enque(Packet* pkt)
{
	double now = Scheduler::instance().clock();
	int m;
	Flow *f = 0;

        if (idle_) {
		/* To account for the period when the queue was empty.  */
                idle_ = 0;
		m = int(edp_.ptc * (now - idletime_));
        } else
                m = 0;

	run_estimator(edp_.bytes ? qnb_ : qnp_, m + 1);

	if (trace_)
		if (edp_.bytes) {
			plot1(qnb_);
		}
		else {
			plot1(qnp_);
		}

	if (flowmgr_) {
		if ((f = flowmgr_->findflow(pkt)) == NULL)
			f = flowmgr_->newflow(pkt);
		if (f)
			f->arrival(pkt);
	}

	++edv_.count;
	edv_.count_bytes += pkt->size_;

	if (edv_.v_ave >= edp_.th_min && qnp_ > 1) { 
		if (edv_.old == 0) {
			edv_.count = 1;
			edv_.count_bytes = pkt->size_;
			edv_.old = 1;
		} else 
			/*
			 * Drop each packet with probability edv.v_prob.
			 */
			if (drop_early(pkt)) {
				log_packet_arrival(pkt);
				log_packet_drop(pkt);
				if (f) {
					if (edv_.v_ave < edp_.th_max)
						f->drop(pkt, 1);
                                        else f->drop(pkt, 0);
				}
				pkt = 0;
			}
	} else {
		edv_.v_prob = 0.0;
		edv_.old = 0;
	}
	/*
	 * If we didn't drop the packet above, send it to the interface,
	 * checking for absolute queue overflow.
	 */
	if (pkt != 0) {
		q_.enque(pkt);
		log_packet_arrival(pkt);
		int metric = edp_.bytes ? qnb_ : qnp_;
		int limit = edp_.bytes ? (qlim_* edp_.mean_pktsize): qlim_;
		if (metric > limit) {
			int victim;
			if (drop_tail_)
				victim = qnp_ - 1;
			else
				victim = Random::integer(qnp_);
				
			pkt = q_.lookup(victim);
			q_.remove(pkt);
			log_packet_drop(pkt);
			if (f)
				f->drop(pkt, 0);
			Packet::free(pkt);
		}
	}
	if (trace_)
		plot();
}

//  ERED (REDCopQueue) Code by Wu-chang Feng <wuchang@eecs.umich.edu>
 
class REDCopQueue : virtual public REDQueue {
public:
	REDCopQueue();
protected:
	virtual void enque(Packet* pkt);
	virtual int drop_early(Packet* pkt);
	double pth_min;
	double pth_max;
	int drops_;
};

static class REDCopMatcher : public Matcher {
public:
	REDCopMatcher() : Matcher("link") {}
	TclObject* match(const char* id) {
		if (strcmp(id, "redcop") == 0)
			return (new REDCopQueue);
		return (0);
	}
} matcher_redcop;

REDCopQueue::REDCopQueue()
{
	link_real("ns_redcop", "pthresh", &pth_min, 0);
	link_real("ns_redcop", "pmaxthresh", &pth_max, 0);
	link_int(0, "drops", &drops_, 1);
}

/*
 * Receive a new packet arriving at the queue.
 * The average queue size is computed.  If the average size
 * exceeds the threshold, then the dropping probability is computed,
 * and the newly-arriving packet is dropped with that probability.
 * The packet is also dropped if the maximum queue size is exceeded.
 */
void REDCopQueue::enque(Packet* pkt)
{
	double now = Scheduler::instance().clock();
	int m;
	Flow *f = 0;

        if (idle_) {
		/* To account for the period when the queue was empty.  */
                idle_ = 0;
		m = int(edp_.ptc * (now - idletime_));
        } else
                m = 0;

	run_estimator(edp_.bytes ? qnb_ : qnp_, m + 1);

	if (trace_)
		if (edp_.bytes) {
			plot1(qnb_);
		}
		else {
			plot1(qnp_);
		}

	if (flowmgr_) {
		if ((f = flowmgr_->findflow(pkt)) == NULL)
			f = flowmgr_->newflow(pkt);
		if (f)
			f->arrival(pkt);
	}

	++edv_.count;
	edv_.count_bytes += pkt->size_;

	if ( ((edv_.v_ave >= edp_.th_min && !(pkt->flags_ & PF_PRI)) ||
	     (edv_.v_ave >= pth_min) )
	    && qnp_ > 1) { 
		if (edv_.old == 0) {
			edv_.count = 1;
			edv_.count_bytes = pkt->size_;
			edv_.old = 1;
		} else 
			/*
			 * Drop each packet with probability edv.v_prob.
			 */
			if (drop_early(pkt)) {
				log_packet_arrival(pkt);
				log_packet_drop(pkt);
				drops_++;
				if (f) {
					if (edv_.v_ave < edp_.th_max)
						f->drop(pkt, 1);
                                        else f->drop(pkt, 0);
				}
				pkt = 0;
			}
	} else {
		edv_.v_prob = 0.0;
		edv_.old = 0;
	}
	/*
	 * If we didn't drop the packet above, send it to the interface,
	 * checking for absolute queue overflow.
	 */
	if (pkt != 0) {
		q_.enque(pkt);
		log_packet_arrival(pkt);
		int metric = edp_.bytes ? qnb_ : qnp_;
		int limit = edp_.bytes ? (qlim_* edp_.mean_pktsize): qlim_;
		if (metric > limit) {
			int victim;
			int count;
			if (drop_tail_) {
				victim = qnp_ - 1;
				pkt = q_.lookup(victim);
			}
			else {
				victim = Random::integer(qnp_);
				count = 0;				
				pkt = q_.lookup(victim);
			}
			q_.remove(pkt);
			log_packet_drop(pkt);
			drops_++;
			if (f)
				f->drop(pkt, 0);
			Packet::free(pkt);
		}
	}
	if (trace_)
		plot();
}

int REDCopQueue::drop_early(Packet* pkt)
{
	if ( ((edv_.v_ave >= edp_.th_max) && !(pkt->flags_ & PF_PRI)) ||
	      (edv_.v_ave >= pth_max) ) {
		edv_.v_prob = 1.0;
	}
	else {
		double p = edv_.v_a * edv_.v_ave + edv_.v_b;
		double count1 = edv_.count;
		if ( edp_.bytes )
			count1 = (double) (edv_.count_bytes/edp_.mean_pktsize);
		p /= edp_.max_p_inv;
		if (edp_.wait ) {
			if (count1 * p < 1)
				p = 0;
			else if (count1 * p < 2)
				p /= (2 - count1 * p);
			else
				p = 1;
		} else if (!edp_.wait ) {
			if (count1 * p < 1)
				p /= (1 - count1 * p);
			else
				p = 1;
		}
		if (edp_.bytes && p < 1 ) {
			p = p * pkt->size_ / edp_.mean_pktsize;
			if (p>1) p=1;
		}
		edv_.v_prob = p;
#ifdef DEBUGBYTES
		if (p > 0)
			printf ("bytes: %6d prob: %g\n", pkt->size_, p);
#endif
	}
	double u = Random::uniform();
	if (u <= edv_.v_prob) {
		edv_.count = 0;
		edv_.count_bytes = 0;
		if (edp_.setbit) 
			pkt->flags_ |= PF_ECN;
		else
			return (1);
	}
	return (0);
}

class FREDCopQueue : virtual public REDQueue {
public:
	FREDCopQueue();
protected:
	virtual void enque(Packet* pkt);
	virtual Packet* deque();
	virtual int drop_early(Packet* pkt);
	void dq_update(Packet* pkt);
	void eq_update(Packet* pkt);
	int find_entry(Packet* pkt);
	int find_empty();
	double pth_min;
	double pth_max;
	int drops_;
	q_stat qstat_[MAXQ];
	int last_strike_entry;
	int strike_thresh_;
	int rlimit_thresh_;
};

static class FREDCopMatcher : public Matcher {
public:
	FREDCopMatcher() : Matcher("link") {}
	TclObject* match(const char* id) {
		if (strcmp(id, "fredcop") == 0)
			return (new FREDCopQueue);
		return (0);
	}
} matcher_fredcop;

FREDCopQueue::FREDCopQueue()
{
	link_real("ns_fredcop", "pthresh", &pth_min, 0);
	link_real("ns_fredcop", "pmaxthresh", &pth_max, 0);
	link_int("ns_fredcop", "rlimit-thresh", &rlimit_thresh_, 0);
	link_int("ns_fredcop", "strike-thresh", &strike_thresh_, 0);
	link_int(0, "drops", &drops_, 1);

	last_strike_entry = -1;	
	for (int i = 0; i < MAXQ; i++)
		qstat_[i].valid = 0;
}

/*
 * Receive a new packet arriving at the queue.
 * The average queue size is computed.  If the average size
 * exceeds the threshold, then the dropping probability is computed,
 * and the newly-arriving packet is dropped with that probability.
 * The packet is also dropped if the maximum queue size is exceeded.
 */
void FREDCopQueue::enque(Packet* pkt)
{
	double now = Scheduler::instance().clock();
	int m, qstat_entry;
	Flow *f = 0;

	qstat_entry = find_entry(pkt);

	if ( (qstat_entry >= 0) &&
	     (qstat_[qstat_entry].rlimit != 0) && 
	     (qstat_[qstat_entry].unmarked >= rlimit_thresh_))
	{
		log_packet_arrival(pkt);
		log_packet_drop(pkt);
		drops_++;
		if (f) {
			if (edv_.v_ave < edp_.th_max) 
				f->drop(pkt, 1);
                        else f->drop(pkt, 0);
			Packet::free(pkt);
			pkt = 0;
		}
		return;
	}

        if (idle_) {
		/* To account for the period when the queue was empty.  */
                idle_ = 0;
		m = int(edp_.ptc * (now - idletime_));
        } else
                m = 0;

	run_estimator(edp_.bytes ? qnb_ : qnp_, m + 1);

	if (trace_)
		if (edp_.bytes) {
			plot1(qnb_);
		}
		else {
			plot1(qnp_);
		}

	if (flowmgr_) {
		if ((f = flowmgr_->findflow(pkt)) == NULL)
			f = flowmgr_->newflow(pkt);
		if (f)
			f->arrival(pkt);
	}

	++edv_.count;
	edv_.count_bytes += pkt->size_;

	if ( ((edv_.v_ave >= edp_.th_min && !(pkt->flags_ & PF_PRI)) ||
	     (edv_.v_ave >= pth_min) )
	    && qnp_ > 1) { 
		if (edv_.old == 0) {
			edv_.count = 1;
			edv_.count_bytes = pkt->size_;
			edv_.old = 1;
		} else 
			/*
			 * Drop each packet with probability edv.v_prob.
			 */
			if (drop_early(pkt)) {
				log_packet_arrival(pkt);
				log_packet_drop(pkt);
				drops_++;
				if (f) {
					if (edv_.v_ave < edp_.th_max)
						f->drop(pkt, 1);
                                        else f->drop(pkt, 0);
				}
				Packet::free(pkt);
				pkt = 0;
			}
	} else {
		edv_.v_prob = 0.0;
		edv_.old = 0;
	}
	/*
	 * If we didn't drop the packet above, send it to the interface,
	 * checking for absolute queue overflow.
	 */
	if (pkt != 0) {
		q_.enque(pkt);
		eq_update(pkt);
		log_packet_arrival(pkt);
		int metric = edp_.bytes ? qnb_ : qnp_;
		int limit = edp_.bytes ? (qlim_* edp_.mean_pktsize): qlim_;
		if (metric > limit) {
			int victim;
			int count;
			if (drop_tail_) {
				victim = qnp_ - 1;
				pkt = q_.lookup(victim);
			}
			else {
				victim = Random::integer(qnp_);
				count = 0;				
				pkt = q_.lookup(victim);
			}
			q_.remove(pkt);
			dq_update(pkt);
			log_packet_drop(pkt);
			drops_++;
			if (f)
				f->drop(pkt, 0);
			Packet::free(pkt);
		}
	}
	if (trace_)
		plot();
}

int FREDCopQueue::drop_early(Packet* pkt)
{
	int qstat_entry;
	if ( ((edv_.v_ave >= edp_.th_max) && !(pkt->flags_ & PF_PRI)) ||
	      (edv_.v_ave >= pth_max) ) {
		qstat_entry = find_entry(pkt);
		if ( (qstat_entry >= 0) &&
		     (qstat_[qstat_entry].unmarked >= 2*rlimit_thresh_) ) {
			// printf("strike: %d\n",pkt->src_);
			if (last_strike_entry == qstat_entry) {
				qstat_[qstat_entry].strikes++;
				if (qstat_[qstat_entry].strikes > strike_thresh_)
					qstat_[qstat_entry].rlimit = 1;
			}
			else {
				qstat_[qstat_entry].strikes = 0;
			}
			edv_.v_prob = 1.0;
			last_strike_entry = qstat_entry;
		}
	}
	else {
		double p = edv_.v_a * edv_.v_ave + edv_.v_b;
		double count1 = edv_.count;
		if ( edp_.bytes )
			count1 = (double) (edv_.count_bytes/edp_.mean_pktsize);
		p /= edp_.max_p_inv;
		if (edp_.wait ) {
			if (count1 * p < 1)
				p = 0;
			else if (count1 * p < 2)
				p /= (2 - count1 * p);
			else
				p = 1;
		} else if (!edp_.wait ) {
			if (count1 * p < 1)
				p /= (1 - count1 * p);
			else
				p = 1;
		}
		if (edp_.bytes && p < 1 ) {
			p = p * pkt->size_ / edp_.mean_pktsize;
			if (p>1) p=1;
		}
		edv_.v_prob = p;
#ifdef DEBUGBYTES
		if (p > 0)
			printf ("bytes: %6d prob: %g\n", pkt->size_, p);
#endif
	}
	double u = Random::uniform();
	if (u <= edv_.v_prob) {
		edv_.count = 0;
		edv_.count_bytes = 0;
		if (edp_.setbit) 
			pkt->flags_ |= PF_ECN;
		else
			return (1);
	}
	return (0);
}

/*
 * Return the next packet in the queue for transmission.
 */
Packet* FREDCopQueue::deque()
{
	Packet* p = q_.deque();
	if (p != 0) {
		idle_ = 0;
		log_packet_departure(p);
		dq_update(p);
	} else {
		idle_ = 1;
		idletime_ = Scheduler::instance().clock();
	}
	return (p);
}

int FREDCopQueue::find_entry(Packet* pkt) {
	for (int i=0; i < MAXQ; i++) {
		if ( qstat_[i].valid && 
		     (pkt->src_ == qstat_[i].src) &&
		     (pkt->dst_ == qstat_[i].dst) &&
		     (pkt->sport_ == qstat_[i].sport) &&
		     (pkt->dport_ == qstat_[i].dport) ) {
			return(i);
		}
	}
	return(-1);
}

int FREDCopQueue::find_empty() {
	for (int i=0; i < MAXQ; i++)
		if ( !(qstat_[i].valid) )
			return(i);
	return(-1);
}

void FREDCopQueue::dq_update(Packet* pkt) {

	int qstat_entry;

	if (pkt == 0) return;

	qstat_entry = find_entry(pkt);
	if (qstat_entry < 0) 
		return;

	if (pkt->flags_ & PF_PRI) {
		qstat_[qstat_entry].marked--;
	} else {
		qstat_[qstat_entry].unmarked--;
	}
	if ( qstat_[qstat_entry].unmarked <= 0) {
		qstat_[qstat_entry].rlimit = 0;
		qstat_[qstat_entry].strikes = 0;
	}
	if ( (qstat_[qstat_entry].unmarked <= 0) &&
	     (qstat_[qstat_entry].marked <= 0) ) {
		qstat_[qstat_entry].valid = 0;
		if (last_strike_entry == qstat_entry)
			last_strike_entry = -1;
	}
}

void FREDCopQueue::eq_update(Packet* pkt) {
	int qstat_entry;

	if (pkt == 0) return;

	qstat_entry = find_entry(pkt);

	if (qstat_entry < 0) {
		qstat_entry = find_empty();
		if (qstat_entry < 0) return;
		qstat_[qstat_entry].valid = 1;
		qstat_[qstat_entry].src = pkt->src_;
		qstat_[qstat_entry].dst = pkt->dst_;
		qstat_[qstat_entry].sport = pkt->sport_;
		qstat_[qstat_entry].dport = pkt->dport_;
		qstat_[qstat_entry].marked = 0;
		qstat_[qstat_entry].unmarked = 0;
		qstat_[qstat_entry].rlimit = 0;
		qstat_[qstat_entry].strikes = 0;
	}

	if (pkt->flags_ & PF_PRI) {
		qstat_[qstat_entry].marked++;
	} else {
		qstat_[qstat_entry].unmarked++;
	}
}

class REDPQueue : virtual public REDQueue {
public:
	REDPQueue();
protected:
	virtual void enque(Packet* pkt);
	virtual int drop_early(Packet* pkt);
	double bias_;
	int drops_;
};

static class REDPMatcher : public Matcher {
public:
	REDPMatcher() : Matcher("link") {}
	TclObject* match(const char* id) {
		if (strcmp(id, "redp") == 0)
			return (new REDPQueue);
		return (0);
	}
} matcher_redp;

REDPQueue::REDPQueue()
{
	link_real("ns_redp", "bias", &bias_, 0);
        link_int(0, "drops", &drops_, 1);
}

/*
 * Receive a new packet arriving at the queue.
 * The average queue size is computed.  If the average size
 * exceeds the threshold, then the dropping probability is computed,
 * and the newly-arriving packet is dropped with that probability.
 * The packet is also dropped if the maximum queue size is exceeded.
 */
void REDPQueue::enque(Packet* pkt)
{
	double now = Scheduler::instance().clock();
	int m;
	Flow *f = 0;

        if (idle_) {
		/* To account for the period when the queue was empty.  */
                idle_ = 0;
		m = int(edp_.ptc * (now - idletime_));
        } else
                m = 0;

	run_estimator(edp_.bytes ? qnb_ : qnp_, m + 1);

	if (trace_)
		if (edp_.bytes) {
			plot1(qnb_);
		}
		else {
			plot1(qnp_);
		}

	if (flowmgr_) {
		if ((f = flowmgr_->findflow(pkt)) == NULL)
			f = flowmgr_->newflow(pkt);
		if (f)
			f->arrival(pkt);
	}

	++edv_.count;
	edv_.count_bytes += pkt->size_;

	if ( edv_.v_ave >= edp_.th_min && qnp_ > 1) { 
		if (edv_.old == 0) {
			edv_.count = 1;
			edv_.count_bytes = pkt->size_;
			edv_.old = 1;
		} else 
			/*
			 * Drop each packet with probability edv.v_prob.
			 */
			if (drop_early(pkt)) {
				log_packet_arrival(pkt);
				log_packet_drop(pkt);
				drops_++;
				if (f) {
					if (edv_.v_ave < edp_.th_max)
						f->drop(pkt, 1);
                                        else f->drop(pkt, 0);
				}
				pkt = 0;
			}
	} else {
		edv_.v_prob = 0.0;
		edv_.old = 0;
	}
	/*
	 * If we didn't drop the packet above, send it to the interface,
	 * checking for absolute queue overflow.
	 */
	if (pkt != 0) {
		q_.enque(pkt);
		log_packet_arrival(pkt);
		int metric = edp_.bytes ? qnb_ : qnp_;
		int limit = edp_.bytes ? (qlim_* edp_.mean_pktsize): qlim_;
		if (metric > limit) {
			int victim;
			int count;
			if (drop_tail_) {
				victim = qnp_ - 1;
				pkt = q_.lookup(victim);
			}
			else {
				victim = Random::integer(qnp_);
				count = 0;				
				pkt = q_.lookup(victim);
			}
			q_.remove(pkt);
			log_packet_drop(pkt);
			drops_++;
			if (f)
				f->drop(pkt, 0);
			Packet::free(pkt);
		}
	}
	if (trace_)
		plot();
}

int REDPQueue::drop_early(Packet* pkt)
{
        if (edv_.v_ave >= edp_.th_max) {
		edv_.v_prob = 1.0;
	}
	else {
		double p = edv_.v_a * edv_.v_ave + edv_.v_b;
		double count1 = edv_.count;
		if ( edp_.bytes )
			count1 = (double) (edv_.count_bytes/edp_.mean_pktsize);
		p /= edp_.max_p_inv;
		if (edp_.wait ) {
			if (count1 * p < 1)
				p = 0;
			else if (count1 * p < 2)
				p /= (2 - count1 * p);
			else
				p = 1;
		} else if (!edp_.wait ) {
			if (count1 * p < 1)
				p /= (1 - count1 * p);
			else
				p = 1;
		}
		if (edp_.bytes && p < 1 ) {
			p = p * pkt->size_ / edp_.mean_pktsize;
			if (p>1) p=1;
		}
		edv_.v_prob = p;
#ifdef DEBUGBYTES
		if (p > 0)
			printf ("bytes: %6d prob: %g\n", pkt->size_, p);
#endif
	}
	double u = Random::uniform();
	if (pkt->flags_ & PF_PRI)
	  u = u*bias_;
	if (u <= edv_.v_prob) {
		edv_.count = 0;
		edv_.count_bytes = 0;
		if (edp_.setbit) 
			pkt->flags_ |= PF_ECN;
		else
			return (1);
	}
	return (0);
}

// #include <stdio.h>
// #include <stdlib.h>
// #include <unistd.h>

// #include "agent.h"
// #include "node.h"

class CBREDPQueue;
class RClass;

// typedef RClass* (RClass::*LinkageMethod)();

class RClass : public Handler, public NsObject {
public:
	RClass();
  //	virtual int command(int argc, const char*const* argv);
	int	bytes_sent_;
	double  update_interval_;
	double	target_bw_;	/* fraction of link bandwidth */
	double	wbwa_;	/* bytes per second allowed this class */
	double	weight_;		/* to plot for debugging */
	double	probability_;
	double	mark_cnt_;
	int	psize_;
	int	marked_;
	int	unmarked_;
	int	algorithm_;
	int	satisfied();
	int	markpacket();
	void	update_probability();
	void	update_status();
	void	log_class_send(Packet* pkt);
	void	log_class_arrival(Packet* pkt);
	void	handle(Event*);
	void	set_timer();
	Event	timer_;
};

static class RClassMatcher : public Matcher {
public:
	RClassMatcher() : Matcher("rclass") {}
	TclObject* match(const char* id) {
		return (new RClass);
	}
} matcher_rclass;

RClass::RClass()
{
	/* Does this line belong anywhere?  
	Tcl& tcl = Tcl::instance();
	*/
	link_real("ns_rclass", "weight", &weight_, 0);
	link_time("ns_rclass", "update-interval", &update_interval_, 0);
	link_bw("ns_rclass", "target-bw", &target_bw_, 0);
	link_real(0, "wbwa", &wbwa_, 1);
	link_int(0, "algorithm", &algorithm_, 1);
	link_int(0, "marked", &marked_, 1);
	link_int(0, "unmarked", &unmarked_, 1);
	link_int(0, "packet-size", &psize_, 1);

	mark_cnt_ = 1;
	wbwa_ = 0.99*target_bw_;
	marked_ = 0;
	unmarked_ = 0;
	psize_ = 1000;
	probability_ = 1;
	bytes_sent_ = 0;
	set_timer();
}

int RClass::satisfied()
{
	double	now = Scheduler::instance().clock();

	if (wbwa_ < target_bw_) 
	  return(0);
	else 
	  return(1);
}

void RClass::update_status()
{
	double	now = Scheduler::instance().clock();
	double bits,bw;
	bits = (double) bytes_sent_ * 8;
	bytes_sent_ = 0;
	bw = bits/update_interval_;
	wbwa_ = weight_*bw + (1-weight_)*wbwa_;
}

void RClass::log_class_arrival(Packet* pkt)
{
	bytes_sent_ += pkt->size_;
	psize_ = pkt->size_;
}

void RClass::log_class_send(Packet* pkt)
{
	if (pkt->flags_ & PF_PRI)
	  marked_++;
	else
	  unmarked_++;
}

int RClass::markpacket()
{
	mark_cnt_ += probability_;

	if (mark_cnt_ > 1) {
		mark_cnt_--;
		return(1);
	}
	else {
		return(0);
	}
}

void RClass::update_probability()
{
  double cwnd,pwnd;
  double now = Scheduler::instance().clock();
  // if (satisfied())
  //	probability_ -= weight_;
  //  else 
  //	probability_ += weight_;

  cwnd = (wbwa_ * 0.01)/(8*psize_);

  pwnd = probability_ * cwnd;
  if (satisfied())
   pwnd -= 1/cwnd;
  else 
   pwnd += 1/cwnd;
  probability_ = pwnd/cwnd;
  if (probability_ > 1) probability_ = 1;
  if (probability_ < 0) probability_ = 0;
  // printf("%g %g %g %g %g\n",now,probability_,wbwa_,target_bw_,cwnd);
  // fflush(stdout);
}

void RClass::set_timer() 
{
	Scheduler::instance().schedule(this, &timer_, update_interval_);
}

void RClass::handle(Event* e)
{
	update_status();
	set_timer();
}

class CBREDCopQueue : virtual public REDCopQueue {
public:	
	CBREDCopQueue();
	virtual int command(int argc, const char*const* argv);
	virtual Packet	*deque(void);
	virtual void	enque(Packet *pkt);
protected:
	virtual void insert_class(RClass*);
	void	plot(RClass* cl, double idle);

	RClass	*classtab_[MAXCLASS];
	RClass	*list_;
};

static class CBREDCopQueueMatcher : public Matcher {
public:
	CBREDCopQueueMatcher() : Matcher("link") {}
	TclObject* match(const char* id) {
		if (strcasecmp(id, "cbredcop") == 0)
			return (new CBREDCopQueue);
		return (0);
	}
} matcher_cbredcop;

CBREDCopQueue::CBREDCopQueue()
	: list_(0)
{
  //	link_int("ns_cbq", "algorithm", &algorithm_, 0);
  //	link_int("ns_cbq", "max-pktsize", &maxpkt_, 1024);

	int i;
	for (i = 0; i < MAXCLASS; ++i)
		classtab_[i] = 0;
}

/*
 * This implements the following tcl commands:
 *  $cbq insert $class
 *  $cbq bind $class $classID
 */
int CBREDCopQueue::command(int argc, const char*const* argv)
{

  	if (argc == 3) {
  		if (strcmp(argv[1], "insert") == 0) {
  			RClass* p = (RClass*)TclObject::lookup(argv[2]);
  			insert_class(p);
  			return (TCL_OK);
  		}
  	} else if (argc == 4) {
  		if (strcmp(argv[1], "bind") == 0) {
  			RClass* p = (RClass*)TclObject::lookup(argv[2]);
  			int id = atoi(argv[3]);
  			/*XXX check id range */
  			classtab_[id] = p;
  			return (TCL_OK);
  		}
  	}
 	return (Link::command(argc, argv));
}

/*
 * Add a class to this cbq object.
 */
void CBREDCopQueue::insert_class(RClass* p)
{
  //	p->next_ = list_;
  //	list_ = p;
}


/* Plot needs to be implemented still. */ 
void CBREDCopQueue::plot(RClass* cl, double idle)
{
	double t = Scheduler::instance().clock();
   	sprintf(trace_->buffer(), "%g\n", t);
	trace_->dump();
}

/*
 * Map arriving packets to the appropriate class, and add the 
 * arriving packet to the queue for that class.
 * If algorithm==1, and that class could send a packet borrowing 
 * from depth at most d, then "stp->depth" is set to no more than d.
 */
void CBREDCopQueue::enque(Packet* pkt)
{
	double	now = Scheduler::instance().clock();
	RClass	*cl;
	if (pkt)
		cl = classtab_[pkt->class_];
	// log_packet_arrival(pkt);

	//
	// note: the following will *SILENTLY DROP* packets which don't
	// belong to a known CBQ class.  this should probably be changed
	// when simulations with unknown traffic are used
	//

	if (cl != 0) {
	  if (cl->markpacket()) {
	    pkt->flags_ |= PF_PRI;
	  }
	  cl->log_class_arrival(pkt);
	} else {
	  // printf("No Class: %d\n", pkt->class_);
	  // fflush(stdout);
	}

	REDCopQueue::enque(pkt);
	return;
}

Packet* CBREDCopQueue::deque(void)
{
	double now = Scheduler::instance().clock();
	Packet* pkt = REDCopQueue::deque();
	RClass	*cl;
	if (pkt)
		cl = classtab_[pkt->class_];	
	if ((cl != 0) && (pkt != 0)) {
	  cl->update_probability();
	  cl->log_class_send(pkt);
	}
	return (pkt);
}

class CBREDPQueue : virtual public REDPQueue {
public:	
	CBREDPQueue();
	virtual int command(int argc, const char*const* argv);
	virtual Packet	*deque(void);
	virtual void	enque(Packet *pkt);
protected:
	virtual void insert_class(RClass*);
	void	plot(RClass* cl, double idle);

	RClass	*classtab_[MAXCLASS];
	RClass	*list_;
};

static class CBREDPQueueMatcher : public Matcher {
public:
	CBREDPQueueMatcher() : Matcher("link") {}
	TclObject* match(const char* id) {
		if (strcasecmp(id, "cbredp") == 0)
			return (new CBREDPQueue);
		return (0);
	}
} matcher_cbredp;

CBREDPQueue::CBREDPQueue()
	: list_(0)
{
  //	link_int("ns_cbq", "algorithm", &algorithm_, 0);
  //	link_int("ns_cbq", "max-pktsize", &maxpkt_, 1024);

	int i;
	for (i = 0; i < MAXCLASS; ++i)
		classtab_[i] = 0;
}

/*
 * This implements the following tcl commands:
 *  $cbq insert $class
 *  $cbq bind $class $classID
 */
int CBREDPQueue::command(int argc, const char*const* argv)
{

  	if (argc == 3) {
  		if (strcmp(argv[1], "insert") == 0) {
  			RClass* p = (RClass*)TclObject::lookup(argv[2]);
  			insert_class(p);
  			return (TCL_OK);
  		}
  	} else if (argc == 4) {
  		if (strcmp(argv[1], "bind") == 0) {
  			RClass* p = (RClass*)TclObject::lookup(argv[2]);
  			int id = atoi(argv[3]);
  			/*XXX check id range */
  			classtab_[id] = p;
  			return (TCL_OK);
  		}
  	}
 	return (Link::command(argc, argv));
}

/*
 * Add a class to this cbq object.
 */
void CBREDPQueue::insert_class(RClass* p)
{
  //	p->next_ = list_;
  //	list_ = p;
}


/* Plot needs to be implemented still. */ 
void CBREDPQueue::plot(RClass* cl, double idle)
{
	double t = Scheduler::instance().clock();
   	sprintf(trace_->buffer(), "%g\n", t);
	trace_->dump();
}

/*
 * Map arriving packets to the appropriate class, and add the 
 * arriving packet to the queue for that class.
 * If algorithm==1, and that class could send a packet borrowing 
 * from depth at most d, then "stp->depth" is set to no more than d.
 */
void CBREDPQueue::enque(Packet* pkt)
{
	double	now = Scheduler::instance().clock();
	RClass	*cl;
	if (pkt)
		cl = classtab_[pkt->class_];
	// log_packet_arrival(pkt);

	//
	// note: the following will *SILENTLY DROP* packets which don't
	// belong to a known CBQ class.  this should probably be changed
	// when simulations with unknown traffic are used
	//

	if (cl != 0) {
	  if (cl->markpacket()) {
	    pkt->flags_ |= PF_PRI;
	  }
	  cl->log_class_arrival(pkt);
	} else {
	  // printf("No Class: %d\n", pkt->class_);
	  // fflush(stdout);
	}

	REDPQueue::enque(pkt);
	return;
}

Packet* CBREDPQueue::deque(void)
{
	double now = Scheduler::instance().clock();
	Packet* pkt = REDPQueue::deque();
	RClass	*cl;
	if (pkt)
		cl = classtab_[pkt->class_];	
	if ((cl != 0) && (pkt != 0)) {
	  cl->log_class_send(pkt);
	  cl->update_probability();
	}
	return (pkt);
}

//  DRED (DREDQueue) Code by Wu-chang Feng <wuchang@eecs.umich.edu>
 
class DREDQueue : virtual public REDQueue {
public:
	DREDQueue();
protected:
	virtual void enque(Packet* pkt);
	virtual int drop_early(Packet* pkt);
	virtual void plot();
	double alpha_;
	double beta_;
	int status_;	// 0=below min_th
			// 1=above min_th, below max_th
			// 2=above max_th
};

static class DREDMatcher : public Matcher {
public:
	DREDMatcher() : Matcher("link") {}
	TclObject* match(const char* id) {
		if (strcmp(id, "dred") == 0)
			return (new DREDQueue);
		return (0);
	}
} matcher_dred;

DREDQueue::DREDQueue()
{
	link_real("ns_dred", "alpha", &alpha_, 0);
	link_real("ns_dred", "beta", &beta_, 0);
}

/*
 * Receive a new packet arriving at the queue.
 * The average queue size is computed.  If the average size
 * exceeds the threshold, then the dropping probability is computed,
 * and the newly-arriving packet is dropped with that probability.
 * The packet is also dropped if the maximum queue size is exceeded.
 */

void DREDQueue::enque(Packet* pkt)
{
	double now = Scheduler::instance().clock();
	int m;
	Flow *f = 0;

        if (idle_) {
		/* To account for the period when the queue was empty.  */
                idle_ = 0;
		m = int(edp_.ptc * (now - idletime_));
        } else
                m = 0;

	run_estimator(edp_.bytes ? qnb_ : qnp_, m + 1);

	if (trace_)
		if (edp_.bytes) {
			plot1(qnb_);
		}
		else {
			plot1(qnp_);
		}

	if (flowmgr_) {
		if ((f = flowmgr_->findflow(pkt)) == NULL)
			f = flowmgr_->newflow(pkt);
		if (f)
			f->arrival(pkt);
	}

	++edv_.count;
	edv_.count_bytes += pkt->size_;

	if (edv_.v_ave >= edp_.th_min && edv_.v_ave < edp_.th_max && 
	    status_ != 1) {
		status_ = 1;
	}

	if (edv_.v_ave >= edp_.th_max && status_ != 2) {
		status_ = 2;
		edp_.max_p_inv = edp_.max_p_inv/alpha_;
		if (edp_.max_p_inv < 1)
			edp_.max_p_inv = 1;
	}
	if (edv_.v_ave <= edp_.th_min && status_ != 0) {
		status_ = 0;
		edp_.max_p_inv = edp_.max_p_inv * beta_;
		if (edp_.max_p_inv > 128)
			edp_.max_p_inv = 128;
	}

	if (edv_.v_ave >= edp_.th_min && qnp_ > 1) { 
		if (edv_.old == 0) {
			edv_.count = 1;
			edv_.count_bytes = pkt->size_;
			edv_.old = 1;
		} else 
			/*
			 * Drop each packet with probability edv.v_prob.
			 */
			if (drop_early(pkt)) {
				log_packet_arrival(pkt);
				log_packet_drop(pkt);
				if (f) {
					if (edv_.v_ave < edp_.th_max)
						f->drop(pkt, 1);
                                        else f->drop(pkt, 0);
				}
				pkt = 0;
			}
	} else {
		edv_.v_prob = 0.0;
		edv_.old = 0;
	}
	/*
	 * If we didn't drop the packet above, send it to the interface,
	 * checking for absolute queue overflow.
	 */
	if (pkt != 0) {
		q_.enque(pkt);
		log_packet_arrival(pkt);
		int metric = edp_.bytes ? qnb_ : qnp_;
		int limit = edp_.bytes ? (qlim_* edp_.mean_pktsize): qlim_;
		if (metric > limit) {
			int victim;
			if (drop_tail_)
				victim = qnp_ - 1;
			else
				victim = Random::integer(qnp_);
				
			pkt = q_.lookup(victim);
			q_.remove(pkt);
			log_packet_drop(pkt);
			if (f)
				f->drop(pkt, 0);
			Packet::free(pkt);
		}
	}
	if (trace_)
		plot();
}

int DREDQueue::drop_early(Packet* pkt)
{
	if (edv_.v_ave >= edp_.th_max)
		edv_.v_prob = 1.0;
	else {
		double p = edv_.v_a * edv_.v_ave + edv_.v_b;
		double count1 = edv_.count;
		if ( edp_.bytes )
			count1 = (double) (edv_.count_bytes/edp_.mean_pktsize);
		p /= edp_.max_p_inv;
		if (edp_.wait ) {
			if (count1 * p < 1)
				p = 0;
			else if (count1 * p < 2)
				p /= (2 - count1 * p);
			else
				p = 1;
		} else if (!edp_.wait ) {
			if (count1 * p < 1)
				p /= (1 - count1 * p);
			else
				p = 1;
		}
		if (edp_.bytes && p < 1 ) {
			p = p * pkt->size_ / edp_.mean_pktsize;
			if (p>1) p=1;
		}
		edv_.v_prob = p;
#ifdef DEBUGBYTES
		if (p > 0)
			printf ("bytes: %6d prob: %g\n", pkt->size_, p);
#endif
	}
	double u = Random::uniform();
	if (u <= edv_.v_prob) {
		edv_.count = 0;
		edv_.count_bytes = 0;
		if (edp_.setbit) 
			pkt->flags_ |= PF_ECN;
		else
			return (1);
	}
	return (0);
}

/*
 * Output the average queue size and the dropping probability. 
 */
void DREDQueue::plot()
{
	double t = Scheduler::instance().clock();
	sprintf(trace_->buffer(), "a %g %g", t, edv_.v_ave);
	trace_->dump();
	sprintf(trace_->buffer(), "p %g %g", t, edv_.v_prob);
	trace_->dump();
	sprintf(trace_->buffer(), "l %g %g", t, edp_.max_p_inv);
	trace_->dump();
}

//  SFRED (SFREDQueue) Code by Wu-chang Feng <wuchang@eecs.umich.edu>

#define SFR_BINS 23
#define PMRTH 1
 
class SFREDQueue : virtual public REDQueue {
public:
	SFREDQueue();
protected:
	virtual void reset();
	void run_estimator(int nqueued, int m, int level, int bin);
	virtual void enque(Packet* pkt);
	virtual Packet* deque();
	virtual int drop_early(Packet* pkt);
	// virtual void plot();
	void eq_update_bins(Packet* pkt);
	void dq_update_bins(Packet* pkt);
        void decrement_bins(Packet* pkt, double increment);
	void update_ave(Packet* pkt, int level);
        int penalize(Packet* pkt);
        int pcheck(Packet* pkt);
	struct edv bins[SFQ_LEVELS][SFR_BINS];
	int pbox_;
	int pbox_count_;
	int allocation_;
	int delta_;
};

static class SFREDMatcher : public Matcher {
public:
	SFREDMatcher() : Matcher("link") {}
	TclObject* match(const char* id) {
		if (strcmp(id, "sfred") == 0)
			return (new SFREDQueue);
		return (0);
	}
} matcher_sfred;

SFREDQueue::SFREDQueue()
{
	link_int("ns_sfred", "delta", &delta_, 0);
	link_int("ns_sfred", "pbox", &pbox_ , 0);
}

void SFREDQueue::reset()
{
	REDQueue::reset();
	edv_.max_p_inv = edp_.max_p_inv;
	edv_.pkts = 0;
	edv_.status = 0;
	edv_.idle = 1;
	edv_.idletime = 0;
	for (int i=0; i < SFR_BINS; i++)
		for (int j; j < SFQ_LEVELS; j++) {
			memcpy(&edv_,(struct edv*) &bins[j][i],sizeof(edv));
		}
	allocation_ = qlim_/SFR_BINS;
	pbox_count_ = 0;
}

unsigned int sfrhash(unsigned int a, unsigned int b, unsigned int c, unsigned int d, int modulus)
{
        unsigned int h;
        h = ((a << 6) + (b << 12) + (c << 18) + (d << 24) ) % modulus;
        return h;
}

int sfrhash0(Packet* pkt, unsigned int fudge)
{
        unsigned int i,j;
        j = (unsigned int) pkt->dport_ * fudge;
        i = sfrhash(pkt->dst_, pkt->sport_, pkt->src_,
                        j, SFR_BINS);
        return(i); 
}

int sfrhash1(Packet* pkt, unsigned int fudge)
{
        unsigned int i,j;
        j = (unsigned int) pkt->sport_ * fudge;
        i = sfrhash(pkt->src_, pkt->dport_, j,
                        pkt->dst_, SFR_BINS);
        return(i);
}
void SFREDQueue::dq_update_bins(Packet* pkt)
{        
   double now = Scheduler::instance().clock();
   for (int i=0; i < SFQ_LEVELS; i++) {
        bins[i][pkt->h[i]].pkts--;
        if (bins[i][pkt->h[i]].pkts <= 0) {
           bins[i][pkt->h[i]].pkts = 0;
	   bins[i][pkt->h[i]].idle = 1;
	   bins[i][pkt->h[i]].idletime = now;
	}
   }
}
   
void SFREDQueue::eq_update_bins(Packet* pkt)
{
        for (int i=0; i < SFQ_LEVELS; i++)
                bins[i][pkt->h[i]].pkts++;
}

void SFREDQueue::decrement_bins(Packet* pkt, double delta)
{
	for (int i=0; i < SFQ_LEVELS; i++) {
		bins[i][pkt->h[i]].max_p_inv -= delta;
		if (bins[i][pkt->h[i]].max_p_inv < 1.00)
			bins[i][pkt->h[i]].max_p_inv = 1.00;
	}
}               

/*
 * Return the next packet in the queue for transmission.
 */
Packet* SFREDQueue::deque()
{
        Packet* pkt = q_.deque();
        double now = Scheduler::instance().clock();
                
        if (pkt != 0) {
                if (pkt->pbox_) {
                  pkt->pbox_ = 0;
                  pbox_count_--;
                  if (pbox_count_ < 0) pbox_count_ = 0;
                }
                else
                        dq_update_bins(pkt); 
                log_packet_departure(pkt);
        }
        return (pkt);
}

void SFREDQueue::run_estimator(int nqueued, int m, int level, int bin)
{
	float f, f_sl, f_old;

	f = bins[level][bin].v_ave;
	f_sl = bins[level][bin].v_slope;
        while (--m >= 1) {
                f_old = f;
                f *= 1.0 - edp_.q_w;
        }
        f_old = f;
        f *= 1.0 - edp_.q_w;
        f += edp_.q_w * nqueued;
	bins[level][bin].v_ave = f;
	bins[level][bin].v_slope = f_sl;
}

void SFREDQueue::update_ave(Packet* pkt, int level) {
	int m;
	double now = Scheduler::instance().clock();

	if (bins[level][pkt->h[level]].idle) {
		bins[level][pkt->h[level]].idle = 0;
		m = int(edp_.ptc * (now - bins[level][pkt->h[level]].idletime));
	}
	else
		m = 0;
	run_estimator(bins[level][pkt->h[level]].pkts,m+1,level,pkt->h[level]);

	bins[level][pkt->h[level]].count++;
	bins[level][pkt->h[level]].count_bytes += pkt->size_;

	if (bins[level][pkt->h[level]].v_ave >= edp_.th_min &&
	    bins[level][pkt->h[level]].v_ave < edp_.th_max &&
	    bins[level][pkt->h[level]].status != 1)
		bins[level][pkt->h[level]].status = 1;
	if (bins[level][pkt->h[level]].v_ave >= edp_.th_max && 
	    bins[level][pkt->h[level]].status != 2) {
                bins[level][pkt->h[level]].status = 2;
                bins[level][pkt->h[level]].max_p_inv = bins[level][pkt->h[level]].max_p_inv - delta_;
                if (bins[level][pkt->h[level]].max_p_inv < 1)
                        bins[level][pkt->h[level]].max_p_inv = 1;
        }
	if (bins[level][pkt->h[level]].v_ave <= edp_.th_min && 
	    bins[level][pkt->h[level]].status != 0) {
                bins[level][pkt->h[level]].status = 0;
                bins[level][pkt->h[level]].max_p_inv = bins[level][pkt->h[level]].max_p_inv + delta_;
                if (bins[level][pkt->h[level]].max_p_inv > 128)
                        bins[level][pkt->h[level]].max_p_inv = 128;
        }

}

/*
 * Receive a new packet arriving at the queue.
 * The average queue size is computed.  If the average size
 * exceeds the threshold, then the dropping probability is computed,
 * and the newly-arriving packet is dropped with that probability.
 * The packet is also dropped if the maximum queue size is exceeded.
 */
void SFREDQueue::enque(Packet* pkt)
{
	double now = Scheduler::instance().clock();
	int m, low;
	Flow *f = 0;

	/* find hashes */
	pkt->h[0] = sfrhash0(pkt,1);
	pkt->h[1] = sfrhash1(pkt,1);

	update_ave(pkt, 0);
	update_ave(pkt, 1);

        if (flowmgr_) {
                if ((f = flowmgr_->findflow(pkt)) == NULL)
                        f = flowmgr_->newflow(pkt);
                if (f)
                        f->arrival(pkt);
        }

        if (penalize(pkt)) {
            double now = Scheduler::instance().clock();
            log_packet_arrival(pkt);
            log_packet_drop(pkt);
            f->drop(pkt,1);
            if (trace_) {
                sprintf(trace_->buffer(),
                       "penalize drop %g  %d  [%d %d %d %d] -> [%d %d] %g %g",
                        now, qlen(), pkt->src_, pkt->sport_, pkt->dst_,
                        pkt->dport_, pkt->h[0], pkt->h[1],
                        bins[0][pkt->h[0]].max_p_inv, 
			bins[1][pkt->h[1]].max_p_inv);
                trace_->dump();
            }
            Packet::free(pkt);
            pkt = 0;
	    return;
        }
	if (bins[0][pkt->h[0]].v_ave > bins[1][pkt->h[1]].v_ave) 
		low = 1; 
	else 
		low = 0;

	if (bins[low][pkt->h[low]].v_ave >= edp_.th_min && bins[low][pkt->h[low]].pkts > 1) {
		if (bins[low][pkt->h[low]].old == 0) {
			bins[low][pkt->h[low]].count = 1;
			bins[low][pkt->h[low]].count_bytes = pkt->size_;
			bins[low][pkt->h[low]].old = 1;
		} else
			if (drop_early(pkt)) {
				log_packet_arrival(pkt);
				log_packet_drop(pkt);
				pkt = 0;
			}
	} else {
		bins[low][pkt->h[low]].v_prob = 0.0;
		bins[low][pkt->h[low]].old = 0;
	}
        /*
         * If we didn't drop the packet above, send it to the interface,
         * checking for absolute queue overflow.
         */
        if (pkt != 0) {
		q_.enque(pkt);
		log_packet_arrival(pkt);
		if (qnp_ > qlim_) {
			pkt = q_.lookup(qnp_ - 1);
			q_.remove(pkt);
			log_packet_drop(pkt);
                        if (f)
                                f->drop(pkt, 0);
                        Packet::free(pkt);
                }
		else {
			if (pkt->pbox_ == 0)
				eq_update_bins(pkt);
			else
				pbox_count_++;
		}
        }
        if (trace_)
                plot();
}	

int SFREDQueue::drop_early(Packet* pkt)
{
	int low;

	if (bins[0][pkt->h[0]].v_ave > bins[1][pkt->h[1]].v_ave) 
		low = 1; 
	else 
		low = 0;

	if (bins[low][pkt->h[low]].v_ave >= edp_.th_max)
		bins[low][pkt->h[low]].v_prob = 1.0;
	else {
		double p = bins[low][pkt->h[low]].v_a * bins[low][pkt->h[low]].v_ave + bins[low][pkt->h[low]].v_b;
		double count1 = bins[low][pkt->h[low]].count;
		if ( edp_.bytes )
			count1 = (double) (bins[low][pkt->h[low]].count_bytes/edp_.mean_pktsize);
		p /= bins[low][pkt->h[low]].max_p_inv;
		if (edp_.wait ) {
			if (count1 * p < 1)
				p = 0;
			else if (count1 * p < 2)
				p /= (2 - count1 * p);
			else
				p = 1;
		} else if (!edp_.wait ) {
			if (count1 * p < 1)
				p /= (1 - count1 * p);
			else
				p = 1;
		}
		if (edp_.bytes && p < 1 ) {
			p = p * pkt->size_ / edp_.mean_pktsize;
		}
		if (p>1) p=1.0;
		bins[low][pkt->h[low]].v_prob = p;
#ifdef DEBUGBYTES
		if (p > 0)
			printf ("bytes: %6d prob: %g\n", pkt->size_, p);
#endif
	}
	double u = Random::uniform();
	if (u <= bins[low][pkt->h[low]].v_prob) {
		bins[low][pkt->h[low]].count = 0;
		bins[low][pkt->h[low]].count_bytes = 0;
		if (edp_.setbit) 
			pkt->flags_ |= PF_ECN;
		else
			return (1);
	}
	return (0);
}

int SFREDQueue::penalize(Packet* pkt)
{
  double now =  Scheduler::instance().clock();

  if (pcheck(pkt)) {
	if (pbox_count_ < pbox_) {
		pkt->pbox_ = 1;
		return(0);
	}
	else {
		decrement_bins(pkt,delta_);
		return(1);
	}
  }
  return(0);
}

int SFREDQueue::pcheck(Packet* pkt)
{
    // Returns 1 if all max_p_inv > 0.98 and all bins are not empty
    // Returns 0 otherwise
    // if ((bins[i][pkt->h[i]].max_p_inv <= PMRTH) || (bins[i][pkt->h[i]].pkts == 0))
    for (int i=0; i < SFQ_LEVELS; i++) {
        if (bins[i][pkt->h[i]].max_p_inv > PMRTH)
            return(0);
    }
    return(1);
}
